const YEARCODE_PATTERN = /^[01]\d{2}$/i;
const TERMCODE_PATTERN = /^[01]\d{2}[2346]$/i;

export class YearCode {
  public readonly centuryOffset: '0' | '1';
  public readonly yearOffset: string;

  public get fromYear(): number {
    const century = 1900 + parseInt(this.centuryOffset, 10) * 100;
    const year = century + parseInt(this.yearOffset, 10) - 1;
    return year;
  }

  public get fall(): TermCode {
    return new TermCode(`${this.toString()}2`);
  }

  public get spring(): TermCode {
    return new TermCode(`${this.toString()}4`);
  }

  public get summer(): TermCode {
    return new TermCode(`${this.toString()}6`);
  }

  public get toYear(): number {
    return this.fromYear + 1;
  }

  constructor(str: string) {
    if (YearCode.isValid(str) === false) {
      throw new Error(`'${str}' is not a valid year code`);
    }

    this.centuryOffset = str.substr(0, 1) as YearCode['centuryOffset'];
    this.yearOffset = str.substr(1, 2) as YearCode['yearOffset'];
  }

  public equals(other: YearCode): boolean {
    return this.toString() === other.toString();
  }

  public comesBefore(other: YearCode): boolean {
    return this.toString() < other.toString();
  }

  public comesAfter(other: YearCode): boolean {
    return this.toString() > other.toString();
  }

  public toJSON() {
    return {
      centuryOffset: this.centuryOffset,
      yearOffset: this.yearOffset,
    };
  }

  public toString() {
    return `${this.centuryOffset}${this.yearOffset}`;
  }

  public static isValid(str: string): boolean {
    return YEARCODE_PATTERN.test(str);
  }

  public static sort(a: YearCode, b: YearCode): -1 | 0 | 1 {
    if (a.comesBefore(b)) {
      return -1;
    }

    if (a.comesAfter(b)) {
      return 1;
    }

    return 0;
  }

  public static fromString(str: string): YearCode {
    return new YearCode(str);
  }
}

export class TermCode {
  public readonly yearCode: YearCode;
  public readonly termId: '2' | '3' | '4' | '6';

  public get termName(): 'fall' | 'spring' | 'summer' {
    switch (this.termId) {
      case '2':
      case '3':
        return 'fall';
      case '4':
        return 'spring';
      case '6':
        return 'summer';
      default:
        throw new Error(`'${this.termId}' is not a valid term id`);
    }
  }

  public get description(): string {
    const capital = this.termName[0].toUpperCase();
    const rest = this.termName.slice(1);
    if (this.termName === 'fall') {
      return `${capital}${rest} ${this.yearCode.fromYear}`;
    } else {
      return `${capital}${rest} ${this.yearCode.toYear}`;
    }
  }

  constructor(str: string) {
    if (TermCode.isValid(str) === false) {
      throw new Error(`'${str}' is not a valid term code`);
    }

    this.yearCode = new YearCode(str.substr(0, 3));
    this.termId = str.substr(3, 1) as TermCode['termId'];
  }

  public equals(other: TermCode): boolean {
    return this.toString() === other.toString();
  }

  public comesBefore(other: TermCode): boolean {
    return this.toString() < other.toString();
  }

  public comesAfter(other: TermCode): boolean {
    return this.toString() > other.toString();
  }

  public toJSON() {
    return {
      yearCode: this.yearCode,
      termId: this.termId,
    };
  }

  public toString(): string {
    return `${this.yearCode}${this.termId}`;
  }

  public static isValid(str: string): boolean {
    return TERMCODE_PATTERN.test(str);
  }

  public static sort(a: TermCode, b: TermCode): -1 | 0 | 1 {
    if (a.comesBefore(b)) {
      return -1;
    }

    if (a.comesAfter(b)) {
      return 1;
    }

    return 0;
  }

  public static fromString(str: string): TermCode {
    return new TermCode(str);
  }
}
